def get_duplicated_unigrams(rdd, tokenizer, threshold):
    """Spark implementation of duplicated unigram detection"""
    return rdd.flatMap(lambda text: [t for t in tokenizer(text)]) \
        .map(lambda token: (token, 1)) \
        .reduceByKey(lambda a, b: a + b) \
        .filter(lambda x: x[1] >= threshold)






import re
from collections import Counter
from typing import List, Set


NEWLINES_RE = re.compile(r"(?:[\n\r] *){2,}")
TOKENIZE_RE = re.compile(r"\W+")


def find_unigram_counts(corpus):
    """Returns a count of unigrams in the corpus"""
    counter = Counter()
    for doc in corpus:
        counter.update(doc)
    return counter


def find_ngram_counts(corpus, n, check_against, use_hashing, join_char):
    """Returns a count of ngrams in the corpus"""
    counter = Counter()
    for doc in corpus:
        counter.update(get_n_grams_w_checking(doc, n, check_against, use_hashing, join_char))
    return counter


def get_n_grams_w_checking(tokens, n, check_against, use_hashing, join_char: str) -> List:
    """Discover n-grams whose n-1-grams exist in the set check_against"""
    assert n > 1, "This won't work for unigrams as no components"

    n_grams = []

    lhs = join_tokens(tokens[0:n - 1], use_hashing, join_char)

    for i in range(len(tokens) - n + 1):
        rhs = join_tokens(tokens[(i + 1):(i + n)], use_hashing, join_char)
        if (lhs in check_against) and (rhs in check_against):
            n_grams.append(join_tokens(tokens[i:(i + n)], use_hashing, join_char))
        lhs = rhs  # Saves calculating next time

    return n_grams


def join_tokens(tokens, hash_values, join_char):
    if hash_values:
        return hash(join_char.join(tokens))
    else:
        return join_char.join(tokens)


def get_n_grams(tokens: List[str], n: int, hash_values: bool, join_char: str) -> List[str]:
    return [join_tokens(tokens[i:i + n], hash_values, join_char) for i in range(len(tokens) - n + 1)]


def calc_resemblance(doc_ngrams: Set, comparison_ngrams: Set) -> float:
    """
    Discover ratio of document n_grams present in the target set
    """
    n_ngrams = len(doc_ngrams)
    n_in_both = len(doc_ngrams.intersection(comparison_ngrams))
    return n_in_both / n_ngrams


def simple_blockizer(text: str) -> List[str]:
    """Splits text at paragraph markers"""
    return [block.strip() for block in NEWLINES_RE.split(text) if len(block) > 0]


def simple_tokenizer(sentence: str) -> List[str]:
    """Splits text at whitespace markers"""
    return [token for token in TOKENIZE_RE.split(sentence) if len(token) > 0]



import pytest

from pyonion.utils import calc_resemblance, get_n_grams_w_checking, simple_tokenizer, simple_blockizer, join_tokens, \
    get_n_grams, find_unigram_counts, find_ngram_counts

resemblance_data = [
    ({'a', 'b', 'c', 'd'}, {'d', 'e', 'f'}, .25),
    ({'a', 'b', 'c', 'd'}, {'e', 'f'}, 0.),
    ({'a', 'b', 'c', 'd'}, {'a', 'b', 'c', 'd', 'e', 'f'}, 1.)
]
n_grams_checking_data = [
    (['the', 'cat', 'sat', 'on', 'the', 'mat'], 4, ['the_cat_sat', 'cat_sat_on', 'on_the_mat'], ['the_cat_sat_on']),
    (['the', 'cat', 'sat', 'on', 'the', 'mat'], 5, ['the_cat_sat_on'], []),
    (['the', 'cat', 'sat', 'on', 'the', 'mat'], 5, ['the_cat_sat_on', 'cat_sat_on_the', 'sat_on_the_mat'],
     ['the_cat_sat_on_the', 'cat_sat_on_the_mat'])
]


@pytest.mark.parametrize("set_a, set_b, r", resemblance_data)
def test_calc_resemblance(set_a, set_b, r):
    assert calc_resemblance(set_a, set_b) == r


@pytest.mark.parametrize("tokens, n, check_against, expected_ngrams", n_grams_checking_data)
def test_get_n_grams_w_checking(tokens, n, check_against, expected_ngrams):
    use_hashing = False
    join_char = '_'
    n_grams = get_n_grams_w_checking(tokens, n, check_against, use_hashing, join_char)
    assert n_grams == expected_ngrams


tokeniser_test_data = [
    ("I'm a test document", ["I", "m", "a", "test", "document"]),
    ("zxzxc", ["zxzxc"]),
    (".£$%", []),
    ("ワタシワアリクツデウ", ["ワタシワアリクツデウ"]),
    ("", [])
]


@pytest.mark.parametrize("doc, expected_tokens", tokeniser_test_data)
def test_simple_tokenizer(doc, expected_tokens):
    assert simple_tokenizer(doc) == expected_tokens


blockizer_test_data = [
    ("""I'm text.
    
    But in blocks!""", ["I'm text.", "But in blocks!"]),
    ("""""", []),
    ("""I use carriage returns\r\n\nand new lines\n \n together""",
     ["I use carriage returns", "and new lines", "together"]),
    ("""This is some text written in blocks
        
        This is a repeated block
        
        This is a clean block""",
     ["This is some text written in blocks", "This is a repeated block", "This is a clean block"]),
    ("This is some text written in blocks\nsome of my blocks only include single returns\n\nThis is a clean block",
     ["This is some text written in blocks\nsome of my blocks only include single returns", "This is a clean block"])
]


@pytest.mark.parametrize("doc, expected_blocks", blockizer_test_data)
def test_simple_blockizer(doc, expected_blocks):
    blocks = simple_blockizer(doc)
    assert blocks == expected_blocks


join_tokens_test_data = [
    (['some', 'tokens', 'here'], '_', False, 'some_tokens_here'),
    (['some', 'tokens', 'here'], '&', False, 'some&tokens&here'),
    (['some', 'tokens', 'here'], '_', True, hash('some_tokens_here')),
    ([], '_', False, ''),
    ([], '_', True, hash(''))
]


@pytest.mark.parametrize("tokens, join_char, hashing, expected_result", join_tokens_test_data)
def test_join_tokens(tokens, join_char, hashing, expected_result):
    assert join_tokens(tokens, hashing, join_char) == expected_result


get_ngrams_test_data = [
    (['the', 'cat', 'sat', 'on', 'the', 'rug'], 3, '_', False,
     ['the_cat_sat', 'cat_sat_on', 'sat_on_the', 'on_the_rug']),
    (['the', 'cat', 'sat', 'on', 'the', 'rug'], 3, '_', True,
     [hash(ngram) for ngram in ['the_cat_sat', 'cat_sat_on', 'sat_on_the', 'on_the_rug']])
]


@pytest.mark.parametrize("tokens, n_gram, join_char, hashing, expected_result", get_ngrams_test_data)
def test_get_n_grams(tokens, n_gram, join_char, hashing, expected_result):
    assert get_n_grams(tokens, n_gram, hashing, join_char) == expected_result


def test_find_unigram_counts():
    test_corpus = [
        ['cat', 'dog', 'i', '', '@', 'cat'],
        ['cat', 'dog', 'cat', 'cat'],
    ]
    counts = find_unigram_counts(test_corpus)

    assert counts['dog'] == 2
    assert counts['cat'] == 5


def test_find_ngram_counts():
    test_corpus = [
        ['cat', 'dog', 'cat', 'i', '', '@', 'cat'],
        ['cat', 'dog', 'cat', 'i', 'cat'],
    ]
    duplicated = {'cat', 'dog'}

    counts = find_ngram_counts(test_corpus, 2, duplicated, False, '_')
    assert counts['cat_dog'] == 2
    assert counts['dog_cat'] == 2
    assert 'cat_i' not in counts


def test_find_ngram_counts_trigram():
    test_corpus = [
        ['cat', 'dog', 'cat', 'i', '', '@', 'cat'],
        ['cat', 'dog', 'cat', 'i', 'cat'],
        ['cat', 'dog', 'mouse', 'i', 'cat'],
    ]
    duplicated = {'cat_dog', 'dog_cat'}

    counts = find_ngram_counts(test_corpus, n=3, check_against=duplicated, use_hashing=False, join_char='_')
    assert counts['cat_dog_cat'] == 2
    assert 'dog_cat_i' not in counts


def test_find_ngram_counts_hashed():
    test_corpus = [
        ['cat', 'dog', 'cat', 'i', '', '@', 'cat'],
        ['cat', 'dog', 'cat', 'i', 'cat'],
    ]
    duplicated = set([hash(token) for token in ['cat', 'dog']])

    counts = find_ngram_counts(test_corpus, 2, duplicated, use_hashing=True, join_char='_')
    assert counts[hash('cat_dog')] == 2
    assert counts[hash('dog_cat')] == 2
    assert hash('cat_i') not in counts



"""
A rough Python implementation of http://corpus.tools/wiki/Onion.

Will be significantly slower, as does not use C++, or the Judy memory arrays, though is suitable for smaller corpora.
On a machine with 64Gb RAM available a corpus of 20 million records containing 2 billion tokens was processed in roughly
24 hours! This includes having to read the corpus from memory multiple times.
"""
import logging

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())  # Don't log by default





from pyonion.remover import DuplicateRemover, CleaningMode, ListCorpusProvider
from pyonion.utils import simple_blockizer, simple_tokenizer


def test_iter_clean_text_in_blocks_all():
    dr = DuplicateRemover(hash_values=False, join_char='_', n_gram=3, duplication_threshold=2)

    corpus = [
        """This is some text written in blocks
        
        This is a repeated block
        
        This is a clean block""",
        """A second piece of text
        
         This is a repeated block"""
    ]
    expected_cleaned = [
        """This is some text written in blocks\n\nThis is a clean block""",
        """A second piece of text"""
    ]
    corpus_provider = ListCorpusProvider(corpus)

    duplicated = dr.find_duplicated_ngrams(corpus_provider)

    clean_corpus = [doc for doc in dr.iter_clean_text_in_blocks(corpus=corpus_provider,
                                                                duplicated_ngrams=duplicated,
                                                                threshold=.5, mode=CleaningMode.ALL)]
    assert clean_corpus == expected_cleaned


def test_iter_clean_text_in_blocks_all_single_return():
    dr = DuplicateRemover(hash_values=False, join_char='_', n_gram=3, duplication_threshold=2)

    corpus = [
        """This is some text written in blocks\nsome of my blocks only include single returns
        
        This is a repeated block
        
        This is a clean block""",
        """A second piece of text
        
         This is a repeated block"""
    ]
    expected_cleaned = [
        "This is some text written in blocks\nsome of my blocks only include single returns\n\nThis is a clean block",
        """A second piece of text"""
    ]
    corpus_provider = ListCorpusProvider(corpus)

    duplicated = dr.find_duplicated_ngrams(corpus_provider)

    clean_corpus = [doc for doc in dr.iter_clean_text_in_blocks(corpus=corpus_provider,
                                                                duplicated_ngrams=duplicated,
                                                                threshold=.5, mode=CleaningMode.ALL)]
    assert clean_corpus == expected_cleaned


def test_iter_clean_text_in_blocks_first():
    dr = DuplicateRemover(hash_values=False, join_char='_', n_gram=3, duplication_threshold=2)

    corpus = [
        """This is some text written in blocks
        
        This is a repeated block
        
        This is a clean block""",
        """A second piece of text
        
         This is a repeated block"""
    ]
    expected_cleaned = [
        """This is some text written in blocks\n\nThis is a repeated block\n\nThis is a clean block""",
        """A second piece of text"""
    ]
    corpus_provider = ListCorpusProvider(corpus)

    duplicated = dr.find_duplicated_ngrams(corpus_provider)

    clean_corpus = [doc for doc in dr.iter_clean_text_in_blocks(corpus=corpus_provider,
                                                                duplicated_ngrams=duplicated,
                                                                threshold=.5, mode=CleaningMode.FIRST)]
    assert clean_corpus == expected_cleaned


def test_iter_clean_text_in_blocks_first_hashed():
    dr = DuplicateRemover(hash_values=True, join_char='_', n_gram=3, duplication_threshold=2)

    corpus = [
        """This is some text written in blocks
        
        This is a repeated block
        
        This is a clean block""",
        """A second piece of text
        
         This is a repeated block"""
    ]
    expected_cleaned = [
        """This is some text written in blocks\n\nThis is a repeated block\n\nThis is a clean block""",
        """A second piece of text"""
    ]
    corpus_provider = ListCorpusProvider(corpus)

    duplicated = dr.find_duplicated_ngrams(corpus_provider)

    clean_corpus = [doc for doc in dr.iter_clean_text_in_blocks(corpus=corpus_provider,
                                                                duplicated_ngrams=duplicated,
                                                                threshold=.5, mode=CleaningMode.FIRST)]
    assert clean_corpus == expected_cleaned



import logging
import re
import time
from abc import ABC
from enum import Enum
from typing import Set, Iterable, List

from .utils import (find_unigram_counts, find_ngram_counts, get_n_grams, calc_resemblance, simple_tokenizer,
                    simple_blockizer)

logger = logging.getLogger(__name__)

BLOCK_JOIN_CHAR = '\n\n'


class CorpusProvider(ABC):
    """A class to provide data to the duplication remover"""

    def __init__(self, tokenizer=None, blockizer=None):
        self.tokenizer = tokenizer if tokenizer is not None else simple_tokenizer
        self.blockizer = blockizer if blockizer is not None else simple_blockizer

    def iter_docs(self) -> Iterable[str]:
        """An iterator over documents"""
        pass

    def iter_tokens(self) -> Iterable[List[str]]:
        """An iterator over tokenized documents"""
        pass

    def iter_blocks(self) -> Iterable[List[str]]:
        """An iterator over document blocks (usually paragraphs)"""
        pass


class ListCorpusProvider(CorpusProvider):
    """A simple corpus provider to repeatedly iterate over a list"""

    def __init__(self, corpus: List[str], tokenizer=None, blockizer=None):
        super().__init__(tokenizer, blockizer)
        self.corpus = corpus

    def iter_docs(self) -> Iterable[str]:
        for doc in self.corpus:
            yield doc

    def iter_tokens(self) -> Iterable[List[str]]:
        for doc in self.corpus:
            yield self.tokenizer(doc)

    def iter_blocks(self) -> Iterable[List[str]]:
        for doc in self.corpus:
            blocks = self.blockizer(doc)
            yield blocks


class FileCorpusProvider(CorpusProvider):
    """Provide a corpus from a text file containing documents in the format:
    <d>Document text<\\d>
    <d>Next
    possibly multiline
    document<\\d>
    This is my own format."""
    DOCUMENT_START_MARKER = '<d>'
    DOCUMENT_END_REGEX = re.compile(r'<\\d>\n?$')

    def __init__(self, filepath: str, tokenizer=None, blockizer=None):
        super().__init__(tokenizer, blockizer)
        self.filepath = filepath

    def _read_docs(self):
        with open(self.filepath, 'r') as f:
            doc = ''
            for line in f.readlines():
                if line.startswith(self.DOCUMENT_START_MARKER):
                    if len(doc) > 0:
                        raise AssertionError("Multiple starts seen")
                    line = line.replace(self.DOCUMENT_START_MARKER, '')
                if re.search(self.DOCUMENT_END_REGEX, line):
                    line = re.sub(self.DOCUMENT_END_REGEX, '', line)
                    doc += line
                    yield doc
                    doc = ''
                else:
                    doc += line

    def iter_docs(self) -> Iterable[str]:
        yield from self._read_docs()

    def iter_tokens(self) -> Iterable[List[str]]:
        for doc in self._read_docs():
            yield self.tokenizer(doc)

    def iter_blocks(self) -> Iterable[List[str]]:
        for doc in self._read_docs():
            blocks = self.blockizer(doc)
            yield blocks


class CleaningMode(Enum):
    FIRST = 1  # Keep the first instance encountered
    ALL = 2  # Remove all duplicated instances


class DuplicateRemover:
    def __init__(self, hash_values=False, join_char='_', n_gram=10, duplication_threshold=2):
        """
        Uses a corpus provider to discover repeated segments of text in the corpus.

        :param hash_values: If true then work using hashed values - extra computation cost but lower memory requirements
        :param join_char: Character used to join n-grams  e.g. [New, York, Times] -> New_York_Times
        :param n_gram: Look for duplicated shingles/ngrams of this length. 10 is a reasonable value.
        :param duplication_threshold: n-grams occurring more than this many times count as duplicates.
        """
        self.hash_values = hash_values
        self.join_char = join_char
        self.n_gram = n_gram
        self.threshold = duplication_threshold

    def find_duplicated_ngrams(self, corpus: CorpusProvider) -> Set[str]:
        """
        Efficiently find duplicated n_grams in a corpus of documents.

        :param corpus: A provider of documents
        :return: A set of strings which represent the discovered ngrams, joined by the concatenation character.
        """
        logger.info("Finding Unigrams")
        duplicated_ngrams = self._find_duplicated_unigrams(corpus)

        for i in range(2, self.n_gram + 1):
            start_time = time.perf_counter()
            logger.info(f"Finding {i}-grams")
            duplicated_ngrams = self._find_duplicated_ngrams(corpus, i, duplicated_ngrams)
            logger.info(f"Found {len(duplicated_ngrams)} duplicated {i}-grams in {time.perf_counter() - start_time:.2f}"
                        f" seconds.")
        return duplicated_ngrams

    def _find_duplicated_unigrams(self, corpus: CorpusProvider):
        """
        Finds unigrams occurring more than N times

        :param corpus: A provider of a tokenized corpus
        :return: The set of duplicated unigrams
        """
        unigram_counts = find_unigram_counts(corpus.iter_tokens())

        if self.hash_values:
            duplicated_unigrams = {hash(unigram) for unigram, count in unigram_counts.items()
                                   if count >= self.threshold}
        else:
            duplicated_unigrams = {unigram for unigram, count in unigram_counts.items() if count >= self.threshold}
        logger.info(f"Found {len(duplicated_unigrams)} unigrams occurring at least {self.threshold} time(s).")

        return duplicated_unigrams

    def _find_duplicated_ngrams(self, corpus: CorpusProvider, n: int, nminusonegrams: Set):
        """
        Finds duplicated n_grams in a memory efficient way

        The process is not computationally, but is memory efficient. n_grams are only counted if both 'halves' of an
        ngram are  present in the nminusonegrams list, for example (New York Times) would only be counted if both
        (New York) and (York Times) were in the set nminusonegrams. This prevents the list of ngrams from becoming
        staggeringly large.

        :param corpus: EITHER a function generating lists of tokenized documents or a list of tokenized documents.
        :param nminusonegrams: A set of n-1 grams to consider. For the return value to be accurate these must appear at
                               least n times in the text.
        :return: A set of ngrams that appear at least n times in the corpus.
        """
        ngram_counts = find_ngram_counts(corpus=corpus.iter_tokens(),
                                         n=n,
                                         check_against=nminusonegrams,
                                         use_hashing=self.hash_values,
                                         join_char=self.join_char)

        duplicated_ngrams = {ngram for ngram, count in ngram_counts.items() if count >= self.threshold}
        logger.info(f"Found {len(duplicated_ngrams)} {n}-grams occurring at least {self.threshold} time(s).")
        return duplicated_ngrams

    def iter_clean_text(self, corpus: CorpusProvider, duplicated_ngrams: Set[str], threshold: float,
                        mode: CleaningMode) -> Iterable[str]:
        """
        Removes documents with a high ratio of duplicated text

        :param corpus: Corpus provider
        :param mode: Either 'first' or 'all'. Behaviour when encountering duplicated text. If 'first' then will keep the
                     first occurrence encountered, else if set to all will remove all occurrences seen.
        :param duplicated_ngrams: Set of duplicated ngrams - only consider these when looking at resemblence.
        :param threshold: If resemblance with duplicated text is above this then remove this document.
        :return: An iterator of cleaned documents.
        """
        if mode is CleaningMode.FIRST:
            yield from self._clean_text_first(corpus, duplicated_ngrams, threshold)

        elif mode is CleaningMode.ALL:
            yield from self._clean_text_all(corpus, duplicated_ngrams, threshold)

    def _clean_text_all(self, corpus: CorpusProvider, duplicated_ngrams, threshold):
        """Remove ALL documents with high resemblance"""
        for document in corpus.iter_tokens():
            doc_ngrams = set(get_n_grams(document, self.n_gram, self.hash_values, self.join_char))
            resemblance = calc_resemblance(doc_ngrams, duplicated_ngrams)

            if resemblance >= threshold:
                yield '', resemblance
            else:
                yield ' '.join(document), resemblance

    def _clean_text_first(self, corpus: CorpusProvider, duplicated_ngrams, threshold):
        """Keep only the first seen version of each document"""
        seen_n_grams = set()
        for document in corpus.iter_tokens():
            doc_ngrams = set(get_n_grams(document, self.n_gram, self.hash_values, self.join_char))
            resemblance = calc_resemblance(doc_ngrams, seen_n_grams)
            seen_n_grams.update(doc_ngrams.intersection(duplicated_ngrams))

            if resemblance >= threshold:
                yield '', resemblance
            else:
                yield ' '.join(document), resemblance

    def iter_clean_text_in_blocks(self, corpus: CorpusProvider, duplicated_ngrams: Set[str], threshold: float,
                                  mode: CleaningMode) -> Iterable[str]:
        """
        Removes low quality blocks from text

        Splits a document into blocks using the provided blockizer (or paragraph splitter), then finds the number of
        duplicated n_grams in that block. If the resemblance between the block and the duplicate set is above the
        threshold that block will be removed. If all text is of low quality will return an empty string.

        :param corpus:
        :param mode: Either 'first' or 'all'. Behaviour when encountering duplicated text. If 'first' then will keep the
                     first occurrence encountered, else if set to all will remove all occurrences seen.
        :param duplicated_ngrams: Set of duplicated ngrams
        :param threshold: Remove blocks with resemblance above this
        :return: A Iterator over cleaned text. If all blocks of text are poor quality will return an empty string.
        """
        if mode is CleaningMode.FIRST:
            yield from self._clean_blocks_first(corpus, duplicated_ngrams, threshold)

        elif mode is CleaningMode.ALL:
            yield from self._clean_blocks_all(corpus, duplicated_ngrams, threshold)
        else:
            raise NotImplementedError(f"Mode {mode} not recognised. Must be either 'first' or 'all'")

    def _clean_blocks_all(self, corpus: CorpusProvider, duplicated_ngrams, threshold):
        for blocks in corpus.iter_blocks():
            clean_blocks = []
            for block in blocks:
                block_tokens = corpus.tokenizer(block)
                block_ngrams = set(get_n_grams(block_tokens, self.n_gram, self.hash_values, self.join_char))
                resemblance = calc_resemblance(block_ngrams, duplicated_ngrams)

                if resemblance < threshold:
                    clean_blocks.append(block)
            yield BLOCK_JOIN_CHAR.join(clean_blocks)

    def _clean_blocks_first(self, corpus: CorpusProvider, duplicated_ngrams, threshold):
        seen_n_grams = set()
        for blocks in corpus.iter_blocks():
            clean_blocks = []
            for block in blocks:
                block_tokens = corpus.tokenizer(block)
                block_ngrams = set(get_n_grams(block_tokens, self.n_gram, self.hash_values, self.join_char))
                resemblance = calc_resemblance(block_ngrams, seen_n_grams)
                seen_n_grams.update(block_ngrams.intersection(duplicated_ngrams))

                if resemblance < threshold:
                    clean_blocks.append(block)
            yield BLOCK_JOIN_CHAR.join(clean_blocks)



from setuptools import setup
import os

here = os.path.abspath(os.path.dirname(__file__))

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

with open(os.path.join(here, 'requirements-dev.txt')) as f:
    dev_reqs = f.read().splitlines()
with open(os.path.join(here, 'requirements-spark.txt')) as f:
    spark_reqs = f.read().splitlines()


setup(
    name="pyonion",
    use_scm_version={
        "local_scheme": "no-local-version",
        "version_scheme": "guess-next-dev"
    },
    setup_requires=['setuptools_scm'],
    author="KAPUK",
    author_email="alex.clibbon@kantar.com",
    description="A minimal implementation of the ONe Instance ONly algorithm",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=['pyonion'],
    install_requires=[],
    extras_require={
        "dev": dev_reqs,
        "spark": spark_reqs
    },
    package_data={
        'pyonion': ['data/*'],
    },
    url="https://github.com/AClibbon/pyonion",
    classifiers=[
        "Programming Language :: Python :: 3",
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
        'Topic :: Text Processing :: Linguistic'
    ],
    python_requires=">=3.7",

)
